import locale
from dataclasses import dataclass
from datetime import date, datetime
from pathlib import Path
from textwrap import dedent
from typing import Optional, cast

import mistune
from citeproc import CitationStylesBibliography, CitationStylesStyle, formatter
from citeproc.source.bibtex import BibTeX
from dataclass_wizard import DatePattern, DumpMeta, YAMLWizard
from dataclass_wizard import errors as dw_errors
from jinja2.filters import do_striptags
from PIL import Image, UnidentifiedImageError
from slugify import slugify
from yaml.composer import ComposerError

from .customistune import create_custom_markdown
from .generator import mistune_plugins
from .typography import typographie


@dataclass
class Settings:
    title: str
    base_url: str
    extra_vars: str
    target_path: Path
    source_path: Path
    statics_path: Path
    templates_path: Path
    csl_path: Path
    without_statics: bool
    without_search: bool
    feed_limit: int


class FrenchTypographyRenderer(mistune.HTMLRenderer):
    """Apply French typographic rules to text."""

    def text(self, text):
        text = text.replace("\\ ", " ")
        return typographie(super().text(text), html=True)

    def block_html(self, html):
        html = html.replace("\\ ", " ")
        return typographie(super().block_html(html), html=True)


class ImgsWithSizesRenderer(FrenchTypographyRenderer):
    """Renders images as <figure>s and add sizes."""

    def __init__(self, escape=True, allow_harmful_protocols=None, article=None):
        super().__init__(escape, allow_harmful_protocols)
        self._article = article

    def paragraph(self, text):
        # In case of a figure, we do not want the (non-standard) paragraph.
        if text.strip().startswith("<figure>"):
            return text
        return super().paragraph(text)

    def image(self, text, url, title=None):
        if self._article.images_path is None:
            print(f"Image with URL `{url}` is discarded.")
            return ""
        full_path = self._article.images_path.resolve().parent / url
        try:
            image = Image.open(full_path)
        except (IsADirectoryError, FileNotFoundError, UnidentifiedImageError):
            print(f"`{full_path}` is not a valid image.")
            return ""
        width, height = image.size
        caption = f"<figcaption>{text}</figcaption>" if text else ""
        full_url = f"{self._article.settings.base_url}{self._article.url}{url}"
        return dedent(
            f"""\
            <figure>
                <a href="{full_url}"
                    title="Cliquer pour une version haute résolution">
                    <img
                        src="{full_url}"
                        width="{width}" height="{height}"
                        loading="lazy"
                        decoding="async"
                        alt="{text}">
                </a>
                {caption}
            </figure>
            """
        )


@dataclass
class Numero(YAMLWizard):
    _id: str
    name: str
    description: str
    metadata: str
    articles: list
    title: str = ""
    title_f: str = ""

    @property
    def date(self):
        articles_dates = []
        for article in self.articles:
            if article.date:
                articles_dates.append(article.date)
            else:
                print(f"Metadata error in article {article.title}: missing date")
        if articles_dates:
            return max(articles_dates)
        else:
            print("Metadata error no article found with a date")
            return date.today()

    def __lt__(self, other: "Numero"):
        if not isinstance(other, Numero):
            return NotImplemented
        return self.date < other.date

    def __post_init__(self):
        self.slug = slugify(self.name)

    def configure_articles(self, yaml_path, settings):
        # Preserves abstract_fr key (vs. abstract-fr) when converting to_yaml()
        DumpMeta(key_transform="SNAKE").bind_to(Article)

        loaded_articles = []
        for article in self.articles:
            article_slug = slugify(article["article"]["title"])
            article_folder = (
                yaml_path.parent / f"{article_slug}-{article['article']['_id']}"
            )
            article_yaml_path = article_folder / f"{article_slug}.yaml"
            try:
                try:
                    loaded_article = cast(
                        Article, Article.from_yaml_file(article_yaml_path)
                    )
                except ComposerError:
                    loaded_article = cast(
                        Article,
                        Article.from_yaml(
                            article_yaml_path.read_text().split("---")[1]
                        ),
                    )
            except dw_errors.ParseError as e:
                print(f"Metadata error in `{article['article']['title']}`:")
                print(e)
                exit(1)
            if not loaded_article.date:
                print(f"Article `{loaded_article.title}` skipped (no date).")
                continue
            if loaded_article.date > datetime.today().date():
                print(
                    f"Article `{loaded_article.title}` skipped "
                    f"(future date: {loaded_article.date})."
                )
                continue
            if not loaded_article.id:
                loaded_article.id = article_slug
            loaded_article.content_md = (
                article_folder / f"{article_slug}.md"
            ).read_text()
            loaded_article.content_bib_path = article_folder / f"{article_slug}.bib"
            loaded_article.images_path = (
                article_folder / "images"
                if (article_folder / "images").exists()
                else None
            )
            title = loaded_article.dossier[0].get("title")
            title_f = loaded_article.dossier[0].get("title_f")
            if title and title_f:
                self.title = title
                self.title_f = title_f
            else:
                print(f"Article {loaded_article.title} is missing `Issue` metadata")
            loaded_article.numero = self
            loaded_article.settings = settings
            loaded_articles.append(loaded_article)

        self.articles = sorted(loaded_articles, reverse=True)

        for article in self.articles:
            article.init_bibliography()
            article.init_search()


@dataclass
class Article(YAMLWizard):
    title: str
    title_f: str
    id: str = ""
    subtitle: str = ""
    subtitle_f: str = ""
    content_md: str = ""
    content_bib_path: Path = ""
    settings: Settings | None = None
    dossier: list | None = None
    date: Optional[DatePattern["%Y/%m/%d"]] = None  # noqa: F722
    authors: list = None
    abstract: list = None
    keywords: list = None
    numero: Numero | None = None

    def __post_init__(self):
        self.slug = slugify(self.title)
        md = mistune.create_markdown(escape=False)
        self.title_f = md(self.title_f).strip()[len("<p>") : -len("</p>")]
        self.subtitle_f = md(self.subtitle_f).strip()[len("<p>") : -len("</p>")]

    def __eq__(self, other):
        return self.id == other.id

    def __lt__(self, other: "Article"):
        if not isinstance(other, Article):
            return NotImplemented
        return self.date < other.date

    @property
    def abstract_fr(self):
        for abstract in self.abstract:
            if abstract.get("text_f") and (
                abstract.get("lang") == "fr" or abstract.get("lang") is None
            ):
                return abstract["text_f"]

    @property
    def abstract_en(self):
        for abstract in self.abstract:
            if abstract.get("text_f") and abstract.get("lang") == "en":
                return abstract["text_f"]

    @property
    def url(self):
        return f"numero/{self.numero.slug}/article/{self.id}/"

    @property
    def content_html(self):
        md = create_custom_markdown(
            renderer=ImgsWithSizesRenderer(escape=False, article=self),
            plugins=mistune_plugins,
            escape=False,
            bibliography=self.bibliography,
        )
        html_content = md(self.content_md)

        return html_content

    def init_bibliography(self):
        bib_source = BibTeX(self.content_bib_path, encoding="utf-8")
        locale_iso, _ = locale.getlocale()
        if locale_iso is None:
            locale_iso = "fr_FR"
        bib_style = CitationStylesStyle(
            self.settings.csl_path, locale=locale_iso.replace("_", "-"), validate=False
        )
        self.bibliography = CitationStylesBibliography(
            bib_style, bib_source, formatter.html
        )

    def init_search(self):
        self.search_data = {
            "title": self.title,
            "url": f"{self.settings.base_url}{self.url}",
            "date": self.date.strftime("%d %B %Y"),
            "content": do_striptags(
                f"{self.content_html} {self.abstract_fr or self.abstract_en or ''}"
            )
            .replace("'", " ")
            .replace("<", "&lt;")
            .replace(">", "&gt;"),
            "numero_url": f"{self.settings.base_url}numero/{self.numero.slug}/",
            "numero_title_f": self.numero.title_f,
            "numero_date": self.numero.date.strftime("%d %B %Y"),
        }


def configure_numero(yaml_path, settings):
    # Preserves abstract_fr key (vs. abstract-fr) when converting to_yaml()
    DumpMeta(key_transform="SNAKE").bind_to(Numero)

    try:
        numero = Numero.from_yaml_file(yaml_path)
    except ComposerError:
        numero = Numero.from_yaml(yaml_path.read_text().split("---")[1])

    numero.configure_articles(yaml_path, settings)
    return numero


@dataclass
class Keyword:
    slug: str
    name: str
    articles: list

    def __eq__(self, other):
        return self.slug == other.slug

    def __lt__(self, other: "Keyword"):
        if not isinstance(other, Keyword):
            return NotImplemented
        len_self = len(self.articles)
        len_other = len(other.articles)
        if len_self == len_other:
            return self.slug > other.slug
        return len_self < len_other


@dataclass
class Author:
    slug: str
    forname: str
    surname: str
    articles: list
    biography: str = ""

    def __str__(self):
        return f"{self.forname} {self.surname}"

    def __eq__(self, other):
        return self.slug == other.slug

    def __lt__(self, other: "Author"):
        if not isinstance(other, Author):
            return NotImplemented
        len_self = len(self.articles)
        len_other = len(other.articles)
        if len_self == len_other:
            return self.slug > other.slug
        return len_self < len_other


def collect_keywords(numeros):
    keywords = {}
    for numero in numeros:
        for article in numero.articles:
            article_keywords = []
            for kwds in article.keywords:
                if kwds.get("list") and kwds.get("lang") == "fr":  # TODO: en?
                    for keyword in kwds.get("list", "").split(", "):
                        keyword_slug = slugify(keyword)
                        if keyword_slug in keywords:
                            keywords[keyword_slug].articles.append(article)
                            kw = keywords[keyword_slug]
                        else:
                            kw = Keyword(
                                slug=keyword_slug, name=keyword, articles=[article]
                            )
                            keywords[keyword_slug] = kw
                        article_keywords.append(kw)
            article.keywords = article_keywords
    return dict(sorted(keywords.items(), key=lambda item: item[1], reverse=True))


def collect_authors(numeros):
    authors = {}
    for numero in numeros:
        for article in numero.articles:
            article_authors = []
            if not article.authors:
                continue
            for athr in article.authors:
                author_forname = athr.get("forname", "")
                author_surname = athr.get("surname", "")
                author_biography = athr.get("biography", "")
                author_name = f"{author_forname} {author_surname}".strip()
                if not author_name:
                    continue
                author_slug = slugify(author_name)
                if author_slug in authors:
                    authors[author_slug].articles.append(article)
                    kw = authors[author_slug]
                else:
                    kw = Author(
                        slug=author_slug,
                        forname=author_forname,
                        surname=author_surname,
                        articles=[article],
                        biography=author_biography,
                    )
                    authors[author_slug] = kw
                article_authors.append(kw)
            article.authors = article_authors
    return dict(sorted(authors.items(), key=lambda item: item[1], reverse=True))



from typing import TYPE_CHECKING, Any, Dict, List, Match, Union

from citeproc import Citation, CitationItem, CitationStylesBibliography
from mistune.core import BlockState

if TYPE_CHECKING:
    from citeproc import CitationStylesBibliography
    from mistune.core import BaseRenderer, InlineState
    from mistune.inline_parser import InlineParser

    from ..customistune import CustomMarkdown


CITATIONS_REFS = r"\[(?P<suppress>\-?)@(?P<ref>.*?)\]"


def clean_ref(citation_ref):
    # TODO: deal with page references (for instance `[@goody_raison_1979, pp.115]`).
    return citation_ref.split(",")[0]


def warn(citation_item):
    print(f"Reference with key '{citation_item.key}' not found in the bibliography.")


def parse_bibliography_wrapper(bibliography: "CitationStylesBibliography"):
    def parse_bibliography(
        inline: "InlineParser", m: Match[str], state: "InlineState"
    ) -> int:
        author_suppressed = clean_ref(m.group("suppress"))  # TODO
        citation_ref = clean_ref(m.group("ref"))
        citations = state.env.get("citations")
        if not citations:
            citations = {}
        if citation_ref not in citations:
            citation = Citation([CitationItem(citation_ref)])
            citations[citation_ref] = citation
            state.env["citations"] = citations
            bibliography.register(citation)
        else:
            citation = citations[citation_ref]
        state.append_token(
            {
                "type": "bibliography_ref",
                "raw": citation_ref,
                "attrs": {"citation": citation},
            }
        )
        return m.end()

    return parse_bibliography


def md_bibliography_hook(
    md: "CustomMarkdown",
    result: Union[str, List[Dict[str, Any]]],
    state: BlockState,
) -> Union[str, List[Dict[str, Any]]]:
    citations = state.env.get("citations")
    if not citations:
        return result

    state = BlockState()
    state.tokens = [{"type": "bibliography"}]
    output = md.render_state(state)
    return result + output  # type: ignore[operator]


def render_bibliography_ref_wrapper(bibliography: "CitationStylesBibliography"):
    def render_bibliography_ref(
        renderer: "BaseRenderer", citation_ref: str, citation: Citation
    ) -> str:
        return "".join(
            f"""\
<a href="#ref_{citation_ref}" id="anchor_{citation_ref}">\
{bibliography.cite(citation, warn)}\
</a>""".split("\n")
        )

    return render_bibliography_ref


def render_bibliography_wrapper(bibliography: "CitationStylesBibliography"):
    def render_bibliography(renderer: "BaseRenderer") -> str:
        html_bibliography = ""

        def clean_item(item):
            # As of 2025, citeproc-py does not support repeated punctuation.
            return str(item).replace("..", ".").replace(".</i>.", ".</i>")

        for citation, item in zip(bibliography.items, bibliography.bibliography()):
            citation_ref = citation.reference.get("key")
            cleaned_item = clean_item(item)
            html_bibliography += f"""
    <li>
        <span id="ref_{citation_ref}">
            {cleaned_item}
            <a href="#anchor_{citation_ref}">↩</a>
        </span>
    </li>
    """.strip()
        return f"""\
<section id="bibliography">
<ol>
    {html_bibliography}
</ol>
</section>
"""

    return render_bibliography


def bibliography(md: "CustomMarkdown") -> None:
    """A mistune plugin to support bibliography with Bibtex.

    :param md: CustomMarkdown instance
    """
    md.inline.register(
        "bibliography",
        CITATIONS_REFS,
        parse_bibliography_wrapper(md.bibliography),
        before="link",
    )
    md.after_render_hooks.append(md_bibliography_hook)

    if md.renderer and md.renderer.NAME == "html":
        md.renderer.register(
            "bibliography_ref", render_bibliography_ref_wrapper(md.bibliography)
        )
        md.renderer.register(
            "bibliography", render_bibliography_wrapper(md.bibliography)
        )



from typing import TYPE_CHECKING, Any, Dict, List, Match, Union

from mistune.core import BlockState

if TYPE_CHECKING:
    from mistune.block_parser import BlockParser
    from mistune.core import BaseRenderer, InlineState
    from mistune.inline_parser import InlineParser
    from mistune.markdown import Markdown


# https://michelf.ca/projects/php-markdown/extra/#footnotes
INLINE_FOOTNOTE = r"\^\[(?P<footnote_inlined>[^\]]*)\]"


def parse_inline_footnote(
    inline: "InlineParser", m: Match[str], state: "InlineState"
) -> int:
    key = m.group("footnote_inlined")
    notes = state.env.get("inline_footnotes")
    if not notes:
        notes = []
    if key not in notes:
        notes.append(key)
        state.env["inline_footnotes"] = notes
    state.append_token(
        {
            "type": "inline_footnote_ref",
            "raw": key,
            "attrs": {"index": notes.index(key) + 1},
        }
    )
    return m.end()


def parse_footnote_item(
    block: "BlockParser", key: str, index: int, state: BlockState
) -> Dict[str, Any]:
    return {
        "type": "inline_footnote_item",
        "children": [{"type": "paragraph", "text": key}],
        "attrs": {"key": key, "index": index},
    }


def md_footnotes_hook(
    md: "Markdown", result: Union[str, List[Dict[str, Any]]], state: BlockState
) -> Union[str, List[Dict[str, Any]]]:
    notes = state.env.get("inline_footnotes")
    if not notes:
        return result

    children = [
        parse_footnote_item(md.block, k, i + 1, state) for i, k in enumerate(notes)
    ]
    state = BlockState()
    state.tokens = [{"type": "inline_footnotes", "children": children}]
    output = md.render_state(state)
    return result + output  # type: ignore[operator]


def render_inline_footnote_ref(renderer: "BaseRenderer", key: str, index: int) -> str:
    i = str(index)
    return f'<sup class="footnote-ref" id="fniref-{i}"><a href="#fni-{i}">{i}</a></sup>'


def render_inline_footnotes(renderer: "BaseRenderer", text: str) -> str:
    return f'<hr>\n<section class="inline-footnotes">\n<ol>\n{text}</ol>\n</section>\n'


def render_inline_footnote_item(
    renderer: "BaseRenderer", text: str, key: str, index: int
) -> str:
    i = str(index)
    paragraph_ending = "</p>"
    back = f'<a href="#fniref-{i}" class="inline-footnote">&#8617;</a>'
    text = f"{text.rstrip()[: -len(paragraph_ending)]}{back}{paragraph_ending}"
    return f'<li id="fni-{i}">{text}</li>\n'


def inline_footnotes(md: "Markdown") -> None:
    """A mistune plugin to support inline footnotes, spec defined at
    https://michelf.ca/projects/php-markdown/extra/#footnotes

    Here is an example:

    .. code-block:: text

        That's some text with a footnote.^[And that's the footnote.]

    It will be converted into HTML:

    .. code-block:: html

        <p>That's some text with a footnote.<sup class="footnote-ref" id="fniref-1"><a href="#fni-1">1</a></sup></p>
        <hr>
        <section class="inline-footnotes">
        <ol>
        <li id="fni-1"><p>And that's the footnote.<a href="#fniref-1" class="inline-footnote">&#8617;</a></p></li>
        </ol>
        </section>

    :param md: Markdown instance
    """
    md.inline.register(
        "inline_footnote",
        INLINE_FOOTNOTE,
        parse_inline_footnote,
        before="link",
    )
    md.after_render_hooks.append(md_footnotes_hook)

    if md.renderer and md.renderer.NAME == "html":
        md.renderer.register("inline_footnote_ref", render_inline_footnote_ref)
        md.renderer.register("inline_footnote_item", render_inline_footnote_item)
        md.renderer.register("inline_footnotes", render_inline_footnotes)



import html.entities
import unicodedata
from dataclasses import dataclass

import regex  # pour le support de "\p{}"


@dataclass
class Caractere:
    unicode: str
    html: str

    def __init__(self, name: str):
        self.unicode = unicodedata.lookup(name)
        codepoint = ord(self.unicode)
        html_name = html.entities.codepoint2name.get(codepoint, f"#{codepoint}")
        self.html = f"&{html_name};"


ESPACE_INSECABLE = Caractere(name="NO-BREAK SPACE")
ESPACE_FINE_INSECABLE = Caractere(name="NARROW NO-BREAK SPACE")


def assemble_regexes(*regexes):
    return "|".join(regexes)


def build_regex(avant, apres):
    # \p{} permet de reconnaître un caractère par sa catégorie Unicode
    # "Zs" est la catégorie "Separator, space".
    return (
        rf"((?P<avant>{avant})"
        + rf"(\p{{Zs}}|{ESPACE_INSECABLE.html})"
        + rf"(?P<apres>{apres}))"
        + r"(?!(.(?!<svg))*<\/svg>)"
    )


RE_ESPACE_FINE_INSECABLE = regex.compile(
    assemble_regexes(
        build_regex(r"\w?", r"[;\?!]"),  # Ponctuations doubles.
        build_regex(
            r"\d", r"([ghj]|min|sec|images|mm|hab|kg|mg|µg|L|km|°C|GHz)(\b|$)"
        ),  # Unités.
        build_regex(r"\d", r"(Mo|Ko|Go|Mb|Kb|Gb)(\b|$)"),  # Tailles de fichiers.
        build_regex(r"\d", r"%"),  # Pourcentages.
        build_regex(r"\d", r"€"),  # Symboles monétaires.
        build_regex(r"\d", r"\d"),  # Séparateurs de milliers.
    )
)


def insere_espaces_fines_insecables(texte):
    return RE_ESPACE_FINE_INSECABLE.sub(
        r"\g<avant>" + ESPACE_FINE_INSECABLE.unicode + r"\g<apres>", texte
    )


RE_ESPACE_INSECABLE = regex.compile(
    assemble_regexes(
        build_regex(r"\w?", r":"),  # Deux points.
        build_regex(r"«", r""),  # Guillemets en chevrons.
        build_regex(r"", r"»"),  # Guillemets en chevrons.
        build_regex(
            rf"\b(\d|{ESPACE_FINE_INSECABLE.html})+", r"(?!\d)\w"
        ),  # Nombre suivi de lettres.
        build_regex(r"(M\.|Mme)", r"\w"),  # Titres (Monsieur, Madame).
    )
)


def insere_espaces_insecables(texte):
    return RE_ESPACE_INSECABLE.sub(
        r"\g<avant>" + ESPACE_INSECABLE.unicode + r"\g<apres>", texte
    )


def encode_espaces_insecables_en_html(texte):
    for caractere in (ESPACE_INSECABLE, ESPACE_FINE_INSECABLE):
        texte = texte.replace(caractere.unicode, caractere.html)
    return texte


def typographie(texte, html=False):
    """
    Utilise les espaces insécables fines ou normales lorsque c’est approprié

    https://fr.wikipedia.org/wiki/Espace_ins%C3%A9cable#En_France
    """
    res = insere_espaces_fines_insecables(insere_espaces_insecables(texte))
    if html:
        res = encode_espaces_insecables_en_html(res)
    return res



from typing import Any, Callable, Dict, Iterable, List, Optional, Union

from citeproc import CitationStylesBibliography
from mistune import RendererRef
from mistune.block_parser import BlockParser
from mistune.core import BaseRenderer, BlockState
from mistune.inline_parser import InlineParser
from mistune.markdown import Markdown
from mistune.plugins import Plugin, PluginRef, import_plugin
from mistune.renderers.html import HTMLRenderer

"""
Custom Markdown class instanciation to be able to attach the
`bibliography` attribute to the class upon creation before plugins
are instanciated because we need that element for the bibliography
plugin which handle citations using citeproc-py.
"""


class CustomMarkdown(Markdown):
    """Markdown instance to convert markdown text into HTML or other formats.
    Here is an example with the HTMLRenderer::

        from mistune import HTMLRenderer

        md = Markdown(renderer=HTMLRenderer(escape=False))
        md('hello **world**')

    :param renderer: a renderer to convert parsed tokens
    :param block: block level syntax parser
    :param inline: inline level syntax parser
    :param plugins: mistune plugins to use
    :param bibliography: a bibliography object from citeproc-py
    """

    def __init__(
        self,
        renderer: Optional[BaseRenderer] = None,
        block: Optional[BlockParser] = None,
        inline: Optional[InlineParser] = None,
        plugins: Optional[Iterable[Plugin]] = None,
        bibliography: Optional[CitationStylesBibliography] = None,
    ):
        if block is None:
            block = BlockParser()

        if inline is None:
            inline = InlineParser()

        self.renderer = renderer
        self.block: BlockParser = block
        self.inline: InlineParser = inline
        self.bibliography: CitationStylesBibliography = bibliography
        self.before_parse_hooks: List[Callable[["Markdown", BlockState], None]] = []
        self.before_render_hooks: List[Callable[["Markdown", BlockState], Any]] = []
        self.after_render_hooks: List[
            Callable[
                ["Markdown", Union[str, List[Dict[str, Any]]], BlockState],
                Union[str, List[Dict[str, Any]]],
            ]
        ] = []

        if plugins:
            for plugin in plugins:
                plugin(self)


def create_custom_markdown(
    escape: bool = True,
    hard_wrap: bool = False,
    renderer: Optional[RendererRef] = "html",
    plugins: Optional[Iterable[PluginRef]] = None,
    bibliography: Optional[CitationStylesBibliography] = None,
) -> CustomMarkdown:
    """Create a Markdown instance based on the given condition.

    :param escape: Boolean. If using html renderer, escape html.
    :param hard_wrap: Boolean. Break every new line into ``<br>``.
    :param renderer: renderer instance, default is HTMLRenderer.
    :param plugins: List of plugins.
    :param bibliography: a bibliography object from citeproc-py

    This method is used when you want to re-use a Markdown instance::

        markdown = create_markdown(
            escape=False,
            hard_wrap=True,
        )
        # re-use markdown function
        markdown('.... your text ...')
    """
    if renderer == "ast":
        # explicit and more similar to 2.x's API
        renderer = None
    elif renderer == "html":
        renderer = HTMLRenderer(escape=escape)

    inline = InlineParser(hard_wrap=hard_wrap)
    real_plugins: Optional[Iterable[Plugin]] = None
    if plugins is not None:
        real_plugins = [import_plugin(n) for n in plugins]
    return CustomMarkdown(
        renderer=renderer,
        inline=inline,
        plugins=real_plugins,
        bibliography=bibliography,
    )
